/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "idlc/c_gen/c_stub_body_gen.h"

namespace {
#include "idlc/c_gen/c_stub_body_gen_cb.h"
}

namespace tidl {

CStubBodyGen::CStubBodyGen(std::shared_ptr<Document> doc)
    : CBodyGeneratorBase(doc) {}

void CStubBodyGen::OnInitGen(std::ofstream& stream) {
  GenVersion(stream);
  GenGNUSourceDefinition(stream);
  GenIncludeDefaultHeaders(stream);
  GenIncludeHeader(stream);
  GenLogTag(stream, std::string("RPC_PORT_STUB"));
  GenLogDefinition(stream);
  GenTypedefStubMethod(stream);
  GenStructures(stream);
  GenInterfaces(stream);
}

void CStubBodyGen::OnFiniGen(std::ofstream& stream) {}

void CStubBodyGen::GenInterfaces(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_INTERFACE)
      continue;

    const Interface &inf = static_cast<const Interface&>(*i);
    GenInterface(stream, inf);
  }
}

void CStubBodyGen::GenInterface(std::ofstream& stream, const Interface& inf) {
  GenInterfaceEnumerations(stream, inf);
  GenInterfaceGlobalVariables(stream, inf);
  GenInterfaceContext(stream, inf);
  GenInterfaceDelegators(stream, inf);
  GenInterfaceMethods(stream, inf);
  GenInterfaceMethodTable(stream, inf);
  GenInterfaceOnConnectedEventCB(stream, inf);
  GenInterfaceOnDisconnectedEventCB(stream, inf);
  GenInterfaceOnReceivedEventCB(stream, inf);
  GenInterfaceAddPrivileges(stream, inf);
  GenInterfaceRegister(stream, inf);
  GenInterfaceUnregister(stream, inf);
  GenInterfaceClientNumberGetter(stream, inf);
}

void CStubBodyGen::GenInterfaceMethods(std::ofstream& stream,
                                       const Interface& inf) {
  for (auto& i : inf.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;

    stream << SmartIndent(GenTemplateString(CB_INTERFACE_METHOD,
          [&]()->std::string {
            return GetInterfaceIdWithNamespace(inf);
          },
          [&]()->std::string {
            return i->GetID();
          },
          [&]()->std::string {
            return GetInterfaceIdWithNamespace(inf);
          },
          [&]()->std::string {
            return GetInterfaceIdWithNamespace(inf);
          },
          [&]()->std::string {
            return GetMethodString(inf, *i);
          }));
  }
}

void CStubBodyGen::GenInterfaceMethodTable(std::ofstream& stream,
                                           const Interface& inf) {
  std::string str;
  int cnt = 0;

  for (auto& i : inf.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;
    str += GenTemplateString(CB_INTERFACE_METHOD_FORMAT,
        [&]()->std::string {
          return GetInterfaceIdWithNamespace(inf) + "_METHOD_" + i->GetID();
        },
        [&]()->std::string {
          return "__" + GetInterfaceIdWithNamespace(inf) + "_method_"
              + i->GetID();
        });
    cnt++;
  }

  if (cnt == 0)
    return;

  stream << NLine(1);
  stream << SmartIndent(GenTemplateString(CB_INTERFACE_METHOD_TABLE,
        [&]()->std::string {
          return GetInterfaceIdWithNamespace(inf);
        },
        [&]()->std::string {
          return str;
        }));
}

void CStubBodyGen::GenInterfaceOnConnectedEventCB(std::ofstream& stream,
                                                  const Interface& inf) {
  stream << SmartIndent(
      ReplaceAll(CB_INTERFACE_ON_CONNECTED, "##",
          GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceOnDisconnectedEventCB(std::ofstream& stream,
                                                     const Interface& inf) {
  stream << SmartIndent(
      ReplaceAll(CB_INTERFACE_ON_DISCONNECTED, "##",
          GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceOnReceivedEventCB(std::ofstream& stream,
                                                 const Interface& inf) {
  stream << SmartIndent(
      ReplaceAll(CB_INTERFACE_ON_RECEIVED, "##",
          GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceRegister(std::ofstream& stream,
                                        const Interface& inf) {
  stream << SmartIndent(ReplaceAll(CB_INTERFACE_REGISTER, "##",
      GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceUnregister(std::ofstream& stream,
                                          const Interface& inf) {
  stream << SmartIndent(ReplaceAll(CB_INTERFACE_UNREGISTER, "##",
      GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceGlobalVariables(std::ofstream& stream,
                                               const Interface& inf) {
  stream << SmartIndent(ReplaceAll(CB_GLOBALS, "##",
      GetInterfaceIdWithNamespace(inf)));
}

std::string CStubBodyGen::GetMethodString(const Interface& inf,
                                          const Declaration& decl) {
  std::string str;
  const char port_setter[] = "rpc_port_$$_set_port($$, callback_port);\n";
  const char setter[] = "$$($$, $$);\n";
  const char do_while_block[] =
      "do {\n" \
      "    rpc_port_parcel_h result;\n" \
      "\n" \
      "    rpc_port_parcel_create(&result);\n" \
      "    rpc_port_parcel_write_int32(result, $$_METHOD_Result);\n" \
      "$$" \
      "    r = rpc_port_parcel_send(result, port);\n" \
      "    rpc_port_parcel_destroy(result);\n" \
      "} while (0);\n";
  const char ternary_operation[] = "## ? ## : \"\"";
  int cnt = 0;
  for (auto& i : decl.GetParameters().GetParams()) {
    if (IsDelegateType(inf, i->GetParameterType().GetBaseType())) {
      str += "rpc_port_" + GetInterfaceIdWithNamespace(inf) + "_" +
          i->GetParameterType().GetBaseType().ToString() + "_h " +
          i->GetID() + " = NULL;" + NLine(1);
    } else if (i->GetParameterType().GetBaseType().IsUserDefinedType() ||
               i->GetParameterType().GetBaseType().ToString() == "list" ||
               i->GetParameterType().GetBaseType().ToString() == "array") {
      str += GetReturnTypeString(i->GetParameterType().GetBaseType()) +
          i->GetID() + " = NULL;" + NLine(1);
    } else {
      str += GetReturnTypeString(i->GetParameterType().GetBaseType()) +
          i->GetID() + ";" + NLine(1);
    }
    if (i->GetParameterType().GetDirection() == ParameterType::Direction::IN)
      cnt++;
  }
  str += NLine(1);

  for (auto& i : decl.GetParameters().GetParams()) {
    if (i->GetParameterType().GetDirection() != ParameterType::Direction::IN)
      continue;

    if (i->GetParameterType().GetBaseType().IsUserDefinedType() ||
        i->GetParameterType().GetBaseType().ToString() == "list" ||
        i->GetParameterType().GetBaseType().ToString() == "array") {
      if (IsDelegateType(inf, i->GetParameterType().GetBaseType())) {
        str += "rpc_port_" + GetInterfaceIdWithNamespace(inf) + "_" +
            i->GetParameterType().GetBaseType().ToString() +
            "_create(&" + i->GetID() + ");" + NLine(1);
        str += "\nif (!" + i->GetID() + ") {\n";
        str += "    _E(\"Failed to create handle\");\n";
        str += "    return -1;\n";
        str += "}\n\n";
        str += GenTemplateString(port_setter,
            [&]()->std::string {
              return GetInterfaceIdWithNamespace(inf) + "_" +
                  i->GetParameterType().GetBaseType().ToString();
            },
            [&]()->std::string {
              return i->GetID();
            });
      } else {
         str += GetConstructorString(i->GetParameterType().GetBaseType(),
          i->GetID());
      }
    }

    str += GenTemplateString(setter,
          [&]()->std::string {
            return GetParcelReadFunctionString(
                i->GetParameterType().GetBaseType(), true);
          },
          [&]()->std::string {
            return "parcel";
          },
          [&]()->std::string {
            if (i->GetParameterType().GetBaseType().IsUserDefinedType() ||
                i->GetParameterType().GetBaseType().ToString() == "list" ||
                i->GetParameterType().GetBaseType().ToString() == "array")
              return "&" + i->GetID() + "->parcelable, " + i->GetID();
            return "&" + i->GetID();
          });
  }

  if (cnt > 0)
    str += NLine(1);

  if (decl.GetMethodType() == Declaration::MethodType::SYNC &&
      decl.GetType().ToString() != "void") {
    str += GetReturnTypeString(decl.GetType()) + "ret = ";
  }

  str += "context->callback." + decl.GetID() + "(context";
  for (auto& i : decl.GetParameters().GetParams()) {
    str += ", ";
    if (i->GetParameterType().GetDirection() != ParameterType::Direction::IN)
      str += "&" + i->GetID();
    else
      str += i->GetID();
  }
  str += ", context->user_data);" + NLine(1);
  if (decl.GetMethodType() == Declaration::MethodType::SYNC) {
    str += GenTemplateString(do_while_block,
        [&]()->std::string {
          return GetInterfaceIdWithNamespace(inf);
        },
        [&]()->std::string {
          std::string s;
          for (auto& i : decl.GetParameters().GetParams()) {
            if (i->GetParameterType().GetDirection() ==
                ParameterType::Direction::IN)
              continue;
            s += GenTemplateString(setter,
                [&]()->std::string {
                  return GetParcelWriteFunctionString(
                      i->GetParameterType().GetBaseType(), true);
                },
                [&]()->std::string {
                  return "result";
                },
                [&]()->std::string {
                  auto& t = i->GetParameterType().GetBaseType();
                  if (t.IsUserDefinedType() || t.ToString() == "list" ||
                      t.ToString() == "array")
                    return "&" + i->GetID() + "->parcelable, " + i->GetID();
                  else if (t.ToString() == "string")
                    return ReplaceAll(ternary_operation, "##", i->GetID());
                  return i->GetID();
                });
          }
          if (decl.GetType().ToString() != "void") {
            s += GenTemplateString(setter,
                [&]()->std::string {
                  return GetParcelWriteFunctionString(decl.GetType(), true);
                },
                [&]()->std::string {
                  return "result";
                },
                [&]()->std::string {
                  if (decl.GetType().IsUserDefinedType() ||
                      decl.GetType().ToString() == "list" ||
                      decl.GetType().ToString() == "array")
                    return "&ret->parcelable, ret";
                  else if (decl.GetType().ToString() == "string")
                    return ReplaceAll(ternary_operation, "##", "ret");
                  return "ret";
                });
          }
          return s;
        });
    }

  for (auto& i : decl.GetParameters().GetParams()) {
    str += NLine(1);
    if (IsDelegateType(inf, i->GetParameterType().GetBaseType())) {
      str += "rpc_port_" + GetInterfaceIdWithNamespace(inf) + "_" +
          i->GetParameterType().GetBaseType().ToString() +
          "_destroy(" + i->GetID() + ");";
    } else {
      str += GetDestructorString(i->GetParameterType().GetBaseType(),
          i->GetID());
    }
  }

  str += GetDestructorString(decl.GetType(), "ret");

  return str;
}

std::string CStubBodyGen::GetAddPrivilegeString(const std::string& id,
                                                const Attribute& attr) {
  std::string str;

  str += GenTemplateString(CB_INTERFACE_PRIVILEGE_BLOCK,
      [&]()->std::string {
        return "rpc_port_stub_add_privilege(__" + id +
            "_stub, \"" + attr.GetValue() + "\")";
      },
      [&]()->std::string {
        return attr.GetValue();
      });
  return str;
}

std::string CStubBodyGen::GetTrustedModeString(const std::string& id,
                                               const Attribute& attr) {
  std::string str;

  str += GenTemplateString(CB_INTERFACE_TRUSTED_MODE_BLOCK,
      [&]()->std::string {
        return "rpc_port_stub_set_trusted(__" + id +
            "_stub, " + attr.GetValue() + ")";
      },
      [&]()->std::string {
        return attr.GetValue();
      });
  return str;
}

void CStubBodyGen::GenInterfaceDelegators(std::ofstream& stream,
                                          const Interface& inf) {
  for (auto& i : inf.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() != Declaration::MethodType::DELEGATE)
      continue;
    GenInterfaceDelegator(stream, GetInterfaceIdWithNamespace(inf), *i);
  }
}

void CStubBodyGen::GenInterfaceDelegator(std::ofstream& stream,
                                         const std::string& id,
                                         const Declaration& decl) {
  GenInterfaceDelegatorDeclaration(stream, id, decl);
  GenInterfaceDelegatorSerializer(stream, id, decl);
  GenInterfaceDelegatorDeserializer(stream, id, decl);
  GenInterfaceDelegatorConstructor(stream, id, decl);
  GenInterfaceDelegatorDestructor(stream, id, decl);
  GenInterfaceDelegatorCloner(stream, id, decl);
  GenInterfaceDelegatorInvoker(stream, id, decl);
  GenInterfaceDelegatorPortSetter(stream, id, decl);
}

void CStubBodyGen::GenInterfaceDelegatorDeclaration(std::ofstream& stream,
                                                    const std::string& id,
                                                    const Declaration& decl) {
  stream << SmartIndent(GenTemplateString(CB_INTERFACE_DELEGATOR_DECL,
        [&]()->std::string {
          return id + "_" + decl.GetID();
        }));
}

void CStubBodyGen::GenInterfaceDelegatorConstructor(std::ofstream& stream,
                                                    const std::string& id,
                                                    const Declaration& decl) {
  stream << SmartIndent(GenTemplateString(
        ReplaceAll(CB_INTERFACE_DELEGATOR_CTOR, "##", id + "_" + decl.GetID()),
        [&]()->std::string {
          return id;
        },
        [&]()->std::string {
          return decl.GetID();
        }));
}

void CStubBodyGen::GenInterfaceDelegatorDestructor(std::ofstream& stream,
                                                   const std::string& id,
                                                   const Declaration& decl) {
  stream << SmartIndent(
      ReplaceAll(CB_INTERFACE_DELEGATOR_DTOR, "##", id + "_" + decl.GetID()));
}

void CStubBodyGen::GenInterfaceDelegatorSerializer(std::ofstream& stream,
                                                   const std::string& id,
                                                   const Declaration& decl) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_DELEGATOR_SERIALIZER, "##", id + "_" + decl.GetID()));
}

void CStubBodyGen::GenInterfaceDelegatorDeserializer(std::ofstream& stream,
                                                     const std::string& id,
                                                     const Declaration& decl) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_DELEGATOR_DESERIALIZER, "##", id + "_" + decl.GetID()));
}

void CStubBodyGen::GenInterfaceDelegatorCloner(std::ofstream& stream,
                                               const std::string& id,
                                               const Declaration& decl) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_DELEGATOR_CLONER, "##", id + "_" + decl.GetID()));
}

void CStubBodyGen::GenInterfaceDelegatorInvoker(std::ofstream& stream,
                                                const std::string& id,
                                                const Declaration& decl) {
  const char parcel[] = "$$(parcel, $$);\n";
  const char ternary_operation[] = "## ? ## : \"\"";
  stream << SmartIndent(GenTemplateString(CB_INTERFACE_DELEGATOR_INVOKER,
        [&]()->std::string {
          return id + "_" + decl.GetID();
        },
        [&]()->std::string {
          std::string str;
          str += "rpc_port_" + id + "_" + decl.GetID() + "_h h";
          for (auto& i : decl.GetParameters().GetParams()) {
            str += ", ";
            str += GetParamTypeString(i->GetParameterType().GetDirection(),
                i->GetParameterType().GetBaseType()) + i->GetID();
          }
          return str;
        },
        [&]()->std::string {
          return id;
        },
        [&]()->std::string {
          std::string str;
          for (auto& i : decl.GetParameters().GetParams()) {
            str += GenTemplateString(parcel,
                [&]()->std::string {
                  return GetParcelWriteFunctionString(
                      i->GetParameterType().GetBaseType(), true);
                },
                [&]()->std::string {
                  auto& t = i->GetParameterType().GetBaseType();
                  if (t.IsUserDefinedType() || t.ToString() == "list" ||
                      t.ToString() == "array")
                    return "&" + i->GetID() + "->parcelable, " + i->GetID();
                  else if (t.ToString() == "string")
                    return ReplaceAll(ternary_operation, "##", i->GetID());
                  return i->GetID();
                });
          }
          return str;
        }));
}

void CStubBodyGen::GenInterfaceDelegatorPortSetter(std::ofstream& stream,
                                                   const std::string& id,
                                                   const Declaration& decl) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_DELEGATOR_PORT_SETTER, "##", id + "_" + decl.GetID()));
}

void CStubBodyGen::GenInterfaceContext(std::ofstream& stream,
                                       const Interface& inf) {
  GenInterfaceContextDeclaration(stream, inf);
  GenInterfaceContextConstructor(stream, inf);
  GenInterfaceContextDestructor(stream, inf);
  GenInterfaceContextFinder(stream, inf);
  GenInterfaceContextTagSetter(stream, inf);
  GenInterfaceContextTagGetter(stream, inf);
  GenInterfaceContextSenderGetter(stream, inf);
}

void CStubBodyGen::GenInterfaceContextDeclaration(std::ofstream& stream,
                                                  const Interface& inf) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_CONTEXT_DECL, "##", GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceContextConstructor(std::ofstream& stream,
                                                  const Interface& inf) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_CONTEXT_CTOR, "##", GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceContextDestructor(std::ofstream& stream,
                                                 const Interface& inf) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_CONTEXT_DTOR, "##", GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceContextFinder(std::ofstream& stream,
                                             const Interface& inf) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_CONTEXT_FINDER, "##", GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceContextTagSetter(std::ofstream& stream,
                                                const Interface& inf) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_CONTEXT_TAG_SETTER, "##", GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceContextTagGetter(std::ofstream& stream,
                                                const Interface& inf) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_CONTEXT_TAG_GETTER, "##", GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenInterfaceContextSenderGetter(std::ofstream& stream,
                                                   const Interface& inf) {
  stream << SmartIndent(ReplaceAll(
      CB_INTERFACE_CONTEXT_GET_SENDER, "##", GetInterfaceIdWithNamespace(inf)));
}

void CStubBodyGen::GenTypedefStubMethod(std::ofstream& stream) {
  stream << CB_STUB_METHOD_TYPE;
}

void CStubBodyGen::GenInterfaceAddPrivileges(std::ofstream& stream,
                                             const Interface& inf) {
  stream << SmartIndent(GenTemplateString(CB_INTERFACE_ADD_PRIVILEGE,
        [&]()->std::string {
          return GetInterfaceIdWithNamespace(inf);
        },
        [&]()->std::string {
          std::string str;
          for (auto& a : inf.GetAttributes().GetAttrs()) {
            if (a->GetKey() == "privilege") {
              str += GetAddPrivilegeString(
                  GetInterfaceIdWithNamespace(inf), *a);
              str += NLine(1);
            } else if (a->GetKey() == "trusted" && a->GetValue() == "true") {
              str += GetTrustedModeString(GetInterfaceIdWithNamespace(inf), *a);
              str += NLine(1);
            }
          }

          if (!str.empty())
            str = SmartIndent("int r;\n\n") + str;
          return str;
        }));
}

void CStubBodyGen::GenInterfaceClientNumberGetter(std::ofstream& stream,
                                                  const Interface& inf) {
  stream << SmartIndent(ReplaceAll(CB_INTERFACE_CLIENT_NUMBER_GETTER, "##",
      GetInterfaceIdWithNamespace(inf)));
}

}  // namespace tidl
